from pathlib import Path

import pytest
import requests

import arapy.cli.copy as copymod
from arapy.core.config import AppPaths, Settings


def _make_settings(tmp_path, profile: str):
    paths = AppPaths(
        cache_dir=tmp_path / profile / "cache",
        state_dir=tmp_path / profile / "state",
        response_dir=tmp_path / profile / "responses",
        app_log_dir=tmp_path / profile / "logs",
    ).ensure()
    return Settings(
        server=profile,
        https_prefix="https://",
        verify_ssl=False,
        timeout=1,
        client_id=f"{profile}-client",
        client_secret=f"{profile}-secret",
        active_profile=profile,
        paths=paths,
    )


def _catalog():
    return {
        "modules": {
            "policyelements": {
                "network-device": {
                    "actions": {
                        "list": {
                            "method": "GET",
                            "paths": ["/api/network-device"],
                            "params": [
                                "filter",
                                "sort",
                                "offset",
                                "limit",
                                "calculate_count",
                            ],
                        },
                        "get": {
                            "method": "GET",
                            "paths": [
                                "/api/network-device/{id}",
                                "/api/network-device/name/{name}",
                            ],
                            "params": ["id", "name"],
                        },
                        "add": {
                            "method": "POST",
                            "paths": ["/api/network-device"],
                            "params": ["name", "ip_address", "radius_secret"],
                        },
                        "update": {
                            "method": "PATCH",
                            "paths": ["/api/network-device/{id}"],
                            "params": ["id", "name", "ip_address", "radius_secret"],
                        },
                        "replace": {
                            "method": "PUT",
                            "paths": ["/api/network-device/{id}"],
                            "params": ["id", "name", "ip_address", "radius_secret"],
                        },
                    }
                }
            }
        }
    }


class _BaseCP:
    last_response_meta = None

    def __init__(self, catalog):
        self.catalog = catalog

    def get_action_definition(self, api_catalog, module, service, action):
        return self.catalog["modules"][module][service]["actions"][action]

    def resolve_action(self, api_catalog, module, service, action, args):
        action_def = self.get_action_definition(api_catalog, module, service, action)
        paths = action_def["paths"]
        if action in {"update", "replace"}:
            return action_def, paths[0].replace("{id}", str(args["id"])), ["id"]
        if action == "get" and args.get("name"):
            return action_def, paths[1].replace("{name}", str(args["name"])), ["name"]
        if action == "get" and args.get("id"):
            return action_def, paths[0].replace("{id}", str(args["id"])), ["id"]
        return action_def, paths[0], []


class _SourceCP(_BaseCP):
    def __init__(self, catalog, items):
        super().__init__(catalog)
        self.items = items

    def list(self, api_catalog, token, args, *, params=None):
        return {"_embedded": {"items": self.items}}

    def get(self, api_catalog, token, args, *, params=None):
        if args.get("name"):
            for item in self.items:
                if item.get("name") == args["name"]:
                    return item
        if args.get("id"):
            for item in self.items:
                if item.get("id") == args["id"]:
                    return item
        err = requests.HTTPError("not found")
        err.response = type("Resp", (), {"status_code": 404})()
        raise err


class _TargetCP(_BaseCP):
    def __init__(self, catalog, matches=None):
        super().__init__(catalog)
        self.matches = matches or {}
        self.add_calls = []
        self.update_calls = []
        self.replace_calls = []

    def list(self, api_catalog, token, args, *, params=None):
        if args.get("filter"):
            return {"_embedded": {"items": list(self.matches.values())[:1]}}
        return {"_embedded": {"items": list(self.matches.values())}}

    def get(self, api_catalog, token, args, *, params=None):
        if args.get("name") and args["name"] in self.matches:
            return self.matches[args["name"]]
        if args.get("id"):
            for item in self.matches.values():
                if item.get("id") == args["id"]:
                    return item
        err = requests.HTTPError("not found")
        err.response = type("Resp", (), {"status_code": 404})()
        raise err

    def add(self, api_catalog, token, args, payload):
        self.add_calls.append({"args": args, "payload": payload})
        return {"id": 3001, **payload}

    def update(self, api_catalog, token, args, payload):
        self.update_calls.append({"args": args, "payload": payload})
        return {"id": args["id"], "name": payload["name"], "radius_secret": ""}

    def replace(self, api_catalog, token, args, payload):
        self.replace_calls.append({"args": args, "payload": payload})
        return {"id": args["id"], **payload}


class _PagedSourceCP(_SourceCP):
    def __init__(self, catalog, items):
        super().__init__(catalog, items)
        self.list_calls = []

    def list(self, api_catalog, token, args, *, params=None):
        self.list_calls.append(dict(params or {}))
        offset = int((params or {}).get("offset", 0))
        limit = int((params or {}).get("limit", 25))
        items = self.items[offset : offset + limit]
        response = {"_embedded": {"items": items}}
        if offset == 0:
            response["count"] = len(self.items)
        return response


def test_handle_copy_command_dry_run_create(monkeypatch, tmp_path, capsys):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "abc123",
            }
        ],
    )
    target_cp = _TargetCP(catalog, {})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "dry_run": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    out = capsys.readouterr().out
    assert "Dry run" in out
    assert report["summary"]["created"] == 1
    assert report["items"][0]["action"] == "create"
    assert report["items"][0]["status"] == "planned"


def test_handle_copy_command_fetches_all_source_pages(monkeypatch, tmp_path):
    import arapy.core.pagination as pagination

    monkeypatch.setattr(pagination, "DEFAULT_PAGE_SIZE", 2)
    catalog = _catalog()
    source_cp = _PagedSourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "one",
            },
            {
                "id": 2,
                "name": "switch-b",
                "ip_address": "10.0.0.2",
                "radius_secret": "two",
            },
            {
                "id": 3,
                "name": "switch-c",
                "ip_address": "10.0.0.3",
                "radius_secret": "three",
            },
        ],
    )
    target_cp = _TargetCP(catalog, {})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "dry_run": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    assert report["summary"]["selected"] == 3
    assert report["summary"]["created"] == 3
    assert source_cp.list_calls == [
        {"limit": 2, "offset": 0, "sort": None},
        {"limit": 2, "offset": 2, "sort": None},
    ]


def test_handle_copy_command_honors_explicit_limit(monkeypatch, tmp_path):
    catalog = _catalog()
    source_cp = _PagedSourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "one",
            },
            {
                "id": 2,
                "name": "switch-b",
                "ip_address": "10.0.0.2",
                "radius_secret": "two",
            },
        ],
    )
    target_cp = _TargetCP(catalog, {})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "limit": "1",
            "dry_run": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    assert report["summary"]["selected"] == 1
    assert report["summary"]["created"] == 1
    assert source_cp.list_calls == [{"limit": 1, "offset": 0, "sort": None}]


def test_handle_copy_command_updates_existing_match_and_restores_secret(
    monkeypatch, tmp_path
):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "abc123",
            }
        ],
    )
    target_cp = _TargetCP(catalog, {"switch-a": {"id": 42, "name": "switch-a"}})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "on_conflict": "update",
            "decrypt": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    assert target_cp.update_calls
    assert target_cp.update_calls[0]["args"]["id"] == 42
    assert target_cp.update_calls[0]["payload"] == {
        "name": "switch-a",
        "ip_address": "10.0.0.1",
        "radius_secret": "abc123",
    }
    assert report["summary"]["updated"] == 1
    assert report["items"][0]["response"]["radius_secret"] == "abc123"


def test_handle_copy_command_uses_cached_catalog_by_default(monkeypatch, tmp_path):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [{"id": 1, "name": "switch-a", "ip_address": "10.0.0.1"}],
    )
    target_cp = _TargetCP(catalog, {})
    catalog_calls = []

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    def get_api_catalog(cp, token, settings, force_refresh=False):
        catalog_calls.append(
            {
                "server": settings.server,
                "token": token,
                "force_refresh": force_refresh,
            }
        )
        return catalog

    copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "dry_run": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=get_api_catalog,
    )

    assert len(catalog_calls) == 2
    assert catalog_calls[0]["server"] == "dev"
    assert catalog_calls[1]["server"] == "prod"
    assert catalog_calls[0]["force_refresh"] is False
    assert catalog_calls[1]["force_refresh"] is False


def test_handle_copy_command_omits_blank_secrets_on_update(monkeypatch, tmp_path):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "",
                "tacacs_secret": "",
            }
        ],
    )
    target_cp = _TargetCP(catalog, {"switch-a": {"id": 42, "name": "switch-a"}})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "on_conflict": "update",
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    assert report["summary"]["updated"] == 1
    assert target_cp.update_calls[0]["payload"] == {
        "name": "switch-a",
        "ip_address": "10.0.0.1",
    }


def test_handle_copy_command_fails_early_for_network_device_create_without_credentials(
    monkeypatch, tmp_path, capsys
):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "",
                "tacacs_secret": "",
            }
        ],
    )
    target_cp = _TargetCP(catalog, {})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    report = copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "dry_run": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    out = capsys.readouterr().out
    assert report["summary"]["created"] == 0
    assert report["summary"]["failed"] == 1
    assert report["items"][0]["status"] == "failed"
    assert (
        "did not include usable RADIUS, TACACS+, or SNMP credentials"
        in report["items"][0]["reason"]
    )
    assert "Failure reasons:" in out
    assert "did not include usable RADIUS, TACACS+, or SNMP credentials" in out
    assert not target_cp.add_calls


def test_handle_copy_command_console_masks_secrets_by_default(
    monkeypatch, tmp_path, capsys
):
    catalog = _catalog()
    source_cp = _SourceCP(
        catalog,
        [
            {
                "id": 1,
                "name": "switch-a",
                "ip_address": "10.0.0.1",
                "radius_secret": "abc123",
            }
        ],
    )
    target_cp = _TargetCP(catalog, {})

    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    def build_client(settings, *, mask_secrets=True):
        return source_cp if settings.server == "dev" else target_cp

    copymod.handle_copy_command(
        {
            "module": "copy",
            "copy_module": "policyelements",
            "copy_service": "network-device",
            "from": "dev",
            "to": "prod",
            "all": True,
            "dry_run": True,
            "console": True,
        },
        settings=_make_settings(tmp_path, "prod"),
        build_client=build_client,
        resolve_auth_token=lambda cp, settings: f"{settings.server}-token",
        get_api_catalog=lambda cp, token, settings, force_refresh=True: catalog,
    )

    out = capsys.readouterr().out
    assert "abc123" not in out
    assert '"radius_secret": ""' in out


def test_handle_copy_command_rejects_missing_selector(monkeypatch, tmp_path):
    monkeypatch.setattr(copymod, "list_profiles", lambda: ["dev", "prod"])
    monkeypatch.setattr(
        copymod,
        "load_settings_for_profile",
        lambda profile: _make_settings(tmp_path, profile),
    )

    with pytest.raises(ValueError, match="Use exactly one selector"):
        copymod.handle_copy_command(
            {
                "module": "copy",
                "copy_module": "policyelements",
                "copy_service": "network-device",
                "from": "dev",
                "to": "prod",
            },
            settings=_make_settings(tmp_path, "prod"),
            build_client=lambda settings, mask_secrets=True: None,
            resolve_auth_token=lambda cp, settings: "token",
            get_api_catalog=lambda cp, token, settings, force_refresh=True: _catalog(),
        )



import string

from hypothesis import given
from hypothesis import settings as hyp_settings
from hypothesis import strategies as st

from arapy.cli.parser import _normalize_flag_name, parse_cli
from arapy.core.config import SECRET_FIELDS
from arapy.core.resolver import query_params_for_action
from arapy.io.output import sanitize_secrets

BOOLEAN_FLAGS = {
    "verbose",
    "version",
    "debug",
    "console",
    "all",
    "decrypt",
    "help",
}

SAFE_VALUE_ALPHABET = string.ascii_letters + string.digits + "_-.:/="
POSITIONAL_ALPHABET = string.ascii_letters + string.digits + "._/"

flag_keys = st.from_regex(r"[a-z][a-z0-9_-]{0,11}", fullmatch=True).filter(
    lambda key: key not in BOOLEAN_FLAGS and not key.startswith("_")
)
flag_values = st.text(alphabet=SAFE_VALUE_ALPHABET, min_size=0, max_size=20)
positionals = st.from_regex(
    r"[A-Za-z0-9][A-Za-z0-9._/]{0,15}", fullmatch=True
)

leaf_values = st.one_of(
    st.none(),
    st.booleans(),
    st.integers(),
    st.text(
        alphabet=st.characters(blacklist_categories=("Cs",)),
        min_size=0,
        max_size=12,
    ),
)
plain_keys = st.from_regex(r"[a-z][a-z0-9_]{0,10}", fullmatch=True).filter(
    lambda key: key not in SECRET_FIELDS
)
nested_values = st.recursive(
    leaf_values,
    lambda children: st.one_of(
        st.lists(children, max_size=4),
        st.dictionaries(
            st.one_of(plain_keys, st.sampled_from(SECRET_FIELDS)),
            children,
            max_size=4,
        ),
    ),
    max_leaves=20,
)


@st.composite
def valid_cli_cases(draw):
    bool_flags = draw(st.lists(st.sampled_from(sorted(BOOLEAN_FLAGS)), unique=True, max_size=4))
    key_values = draw(
        st.dictionaries(flag_keys, flag_values, max_size=4)
    )
    args = ["arapy"]
    args.extend(f"--{flag}" for flag in bool_flags)
    args.extend(f"--{key}={value}" for key, value in key_values.items())
    positional_args = draw(st.lists(positionals, max_size=4))
    args.extend(positional_args)
    return args, bool_flags, key_values, positional_args


@st.composite
def completion_cli_cases(draw):
    long_flags = draw(
        st.lists(flag_keys, unique=True, max_size=4)
    )
    short_flags = draw(
        st.lists(st.sampled_from(list(string.ascii_lowercase)), unique=True, max_size=4)
    )
    positional_args = draw(st.lists(positionals, max_size=4))
    args = ["arapy", "--_complete"]
    args.extend(f"--{flag}" for flag in long_flags)
    args.extend(f"-{flag}" for flag in short_flags)
    args.extend(positional_args)
    return args, positional_args


def _assert_secret_masking(original, sanitized):
    if isinstance(original, dict):
        assert isinstance(sanitized, dict)
        assert set(original) == set(sanitized)
        for key, value in original.items():
            if key in SECRET_FIELDS:
                assert sanitized[key] == ""
            else:
                _assert_secret_masking(value, sanitized[key])
        return

    if isinstance(original, list):
        assert isinstance(sanitized, list)
        assert len(original) == len(sanitized)
        for left, right in zip(original, sanitized):
            _assert_secret_masking(left, right)
        return

    assert original == sanitized


class FakeCP:
    def get_action_definition(self, api_catalog, module, service, action):
        return {"params": ["calculate_count"]}


@hyp_settings(max_examples=150, deadline=None)
@given(valid_cli_cases())
def test_parse_cli_property_valid_inputs(case):
    argv, bool_flags, key_values, positional_args = case

    parsed = parse_cli(argv)

    for flag in bool_flags:
        assert parsed[_normalize_flag_name(flag)] is True

    for key, value in key_values.items():
        assert parsed[_normalize_flag_name(key)] == value

    if len(positional_args) >= 1:
        assert parsed["module"] == positional_args[0]
    if len(positional_args) >= 2:
        assert parsed["service"] == positional_args[1]
    if len(positional_args) >= 3:
        assert parsed["action"] == positional_args[2]


@hyp_settings(max_examples=150, deadline=None)
@given(completion_cli_cases())
def test_parse_cli_property_completion_ignores_unknown_flags(case):
    argv, positional_args = case

    parsed = parse_cli(argv)

    assert parsed["_complete"] is True
    if len(positional_args) >= 1:
        assert parsed["module"] == positional_args[0]
    if len(positional_args) >= 2:
        assert parsed["service"] == positional_args[1]
    if len(positional_args) >= 3:
        assert parsed["action"] == positional_args[2]


@hyp_settings(max_examples=150, deadline=None)
@given(
    secret_key=st.sampled_from(SECRET_FIELDS),
    secret_value=nested_values,
    other_value=nested_values,
)
def test_sanitize_secrets_property_masks_recursive_secret_fields(
    secret_key, secret_value, other_value
):
    payload = {secret_key: secret_value, "other": other_value}

    sanitized = sanitize_secrets(payload, mask_secrets=True)

    _assert_secret_masking(payload, sanitized)


@hyp_settings(max_examples=150, deadline=None)
@given(raw_value=st.one_of(st.booleans(), flag_values))
def test_query_params_property_calculate_count_is_lowercase_boolean_string(raw_value):
    parsed = query_params_for_action(
        FakeCP(),
        {},
        {
            "module": "m",
            "service": "s",
            "action": "list",
            "calculate_count": raw_value,
        },
        "list",
    )

    assert parsed["calculate_count"] in {"true", "false"}

    if isinstance(raw_value, str):
        enabled = raw_value.strip().lower() in {"1", "true", "yes", "on"}
    else:
        enabled = bool(raw_value)
    assert parsed["calculate_count"] == ("true" if enabled else "false")



from __future__ import annotations

import html
import json
import logging
import os
import re
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from arapy.core.config import Settings, load_settings

log = logging.getLogger(__name__)

OAUTH_ENDPOINTS: dict[str, str] = {
    "oauth": "/api/oauth",
    "oauth-me": "/api/oauth/me",
    "oauth-privileges": "/api/oauth/privileges",
}

_PLACEHOLDER_RE = re.compile(r"\{([^}]+)\}")
_LIST_QUERY_PARAMS = ("filter", "sort", "offset", "limit", "calculate_count")


@dataclass(frozen=True)
class EndpointCacheConfig:
    ttl_seconds: int = 24 * 3600
    cache_filename: str = "api_endpoints_cache.json"


def _camel_to_kebab(name: str) -> str:
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1-\2", name)
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1-\2", s1)
    return s2.replace("_", "-").lower()


def _clean_apigility_route(route: str) -> str:
    route = re.sub(r"\[.*?\]", "", route)
    return route.rstrip("/") or "/"


def _ensure_api_prefix(path: str) -> str:
    path = path.strip()
    if not path:
        return "/api"
    if path.startswith("http://") or path.startswith("https://"):
        match = re.search(r"https?://[^/]+(?P<path>/.*)$", path)
        path = match.group("path") if match else path
    if not path.startswith("/"):
        path = "/" + path
    if path.startswith("/api/") or path == "/api":
        return path.rstrip("/") or "/api"
    return ("/api" + path).rstrip("/") or "/api"


def _module_to_cli(module_name: str) -> str:
    base = re.sub(r"-v\d+$", "", module_name)
    return base.replace("_", "-").lower()


def _is_json_content(text: str) -> bool:
    stripped = text.lstrip()
    return stripped.startswith("{") or stripped.startswith("[")


def _extract_module_doc_paths_from_api_docs_html(html: str) -> list[str]:
    modules = set()
    for match in re.findall(r"(/api-docs/[A-Za-z0-9_-]+-v\d+)", html):
        modules.add(match)
    for match in re.findall(r"(api-docs/[A-Za-z0-9_-]+-v\d+)", html):
        modules.add("/" + match)
    return sorted(modules)


def _extract_modules_from_api_docs(text: str) -> list[str]:
    if _is_json_content(text):
        try:
            data = json.loads(text)
        except Exception:
            return []
        out: list[str] = []
        if isinstance(data, dict):
            apis = data.get("apis")
            if isinstance(apis, list):
                for item in apis:
                    if not isinstance(item, dict):
                        continue
                    path = item.get("path")
                    if isinstance(path, str):
                        match = re.search(r"(/api-docs/[A-Za-z0-9_-]+-v\d+)", path)
                        if match:
                            out.append(match.group(1))
        return sorted(set(out))
    return _extract_module_doc_paths_from_api_docs_html(text)


def _extract_placeholders(path: str) -> list[str]:
    return _PLACEHOLDER_RE.findall(path)


def _normalize_template_placeholders(path: str) -> str:
    placeholders = _extract_placeholders(path)
    if len(placeholders) == 1:
        name = placeholders[0]
        if name == "id" or name.endswith("_id"):
            return path.replace("{" + name + "}", "{id}")
    return path


def _dedupe_keep_order(items: list[str]) -> list[str]:
    seen: set[str] = set()
    out: list[str] = []
    for item in items:
        if item not in seen:
            seen.add(item)
            out.append(item)
    return out


def _normalize_param_name(name: str) -> str:
    return "id" if name == "id" or name.endswith("_id") else name


def _clean_text(value: Any) -> str | None:
    if not isinstance(value, str):
        return None
    cleaned = html.unescape(value).strip()
    if "<" in cleaned and ">" in cleaned:
        cleaned = re.sub(r"(?i)<br\s*/?>", "\n", cleaned)
        cleaned = re.sub(r"(?i)<(p|div|tr|li|table|thead|tbody|ul|ol|h\d)\b[^>]*>", "\n", cleaned)
        cleaned = re.sub(r"(?i)</(p|div|tr|li|table|thead|tbody|ul|ol|h\d)>", "\n", cleaned)
        cleaned = re.sub(r"(?i)</(td|th)>", " | ", cleaned)
        cleaned = re.sub(r"(?i)<[^>]+>", "", cleaned)

    normalized_lines: list[str] = []
    for raw_line in cleaned.splitlines():
        line = " ".join(raw_line.strip().split())
        if not line:
            continue
        line = re.sub(r"\s+\|\s+", " | ", line)
        line = re.sub(r"\s+([,.;:!?])", r"\1", line)
        line = re.sub(r'"\s*([^"]*?)\s*"', lambda match: f'"{match.group(1).strip()}"', line)
        line = re.sub(r'([{\[(:])\s+', r"\1", line)
        line = re.sub(r'\s+([}\]):])', r"\1", line)
        line = re.sub(r"\s+\|$", "", line)
        normalized_lines.append(line)

    cleaned = "\n".join(normalized_lines)
    return cleaned or None


def _dedupe_texts(items: list[str]) -> list[str]:
    return _dedupe_keep_order([item for item in (_clean_text(x) for x in items) if item])


def _type_label_for_schema(schema: dict[str, Any]) -> str:
    if not isinstance(schema, dict):
        return "object"

    ref = schema.get("$ref")
    if isinstance(ref, str) and ref:
        return ref

    schema_type = schema.get("type") or schema.get("dataType")
    if isinstance(schema_type, str):
        if schema_type.lower() == "array":
            items = schema.get("items")
            if isinstance(items, dict):
                return f"array[{_type_label_for_schema(items)}]"
        return schema_type

    nested = schema.get("schema")
    if isinstance(nested, dict):
        return _type_label_for_schema(nested)

    return "object"


def _model_required_names(model: dict[str, Any]) -> set[str]:
    required = {
        str(name)
        for name in model.get("required", []) or []
        if isinstance(name, str) and name
    }
    properties = model.get("properties")
    if isinstance(properties, dict):
        for name, prop in properties.items():
            if isinstance(prop, dict) and prop.get("required") is True:
                required.add(str(name))
    return required


def _example_for_schema(
    schema: dict[str, Any], models: dict[str, Any], seen: set[str] | None = None
):
    if seen is None:
        seen = set()

    if not isinstance(schema, dict):
        return {}

    ref = schema.get("$ref")
    if isinstance(ref, str) and ref:
        return _example_for_model(models, ref, seen=seen)

    nested = schema.get("schema")
    if isinstance(nested, dict):
        return _example_for_schema(nested, models, seen=seen)

    enum = schema.get("enum")
    if isinstance(enum, list) and enum:
        return enum[0]

    default_value = schema.get("defaultValue")
    if default_value not in (None, ""):
        return default_value

    schema_type = str(schema.get("type") or schema.get("dataType") or "").lower()
    if schema_type in {"int", "integer", "long"}:
        return 0
    if schema_type in {"number", "float", "double"}:
        return 0
    if schema_type in {"bool", "boolean"}:
        return True
    if schema_type == "array":
        items = schema.get("items")
        if isinstance(items, dict):
            return [_example_for_schema(items, models, seen=set(seen))]
        return []
    if schema_type == "object":
        properties = schema.get("properties")
        if isinstance(properties, dict):
            return {
                str(name): _example_for_schema(prop, models, seen=set(seen))
                for name, prop in properties.items()
                if isinstance(prop, dict)
            }
        return {}
    return ""


def _example_for_model(
    models: dict[str, Any], model_name: str, seen: set[str] | None = None
):
    if seen is None:
        seen = set()
    if not model_name or model_name in seen:
        return {}
    seen.add(model_name)

    model = models.get(model_name)
    if not isinstance(model, dict):
        return {}

    example: dict[str, Any] = {}

    extends = model.get("extends")
    if isinstance(extends, dict):
        ref = extends.get("$ref")
        if isinstance(ref, str) and ref:
            parent = _example_for_model(models, ref, seen=set(seen))
            if isinstance(parent, dict):
                example.update(parent)

    properties = model.get("properties")
    if isinstance(properties, dict):
        for name, prop in properties.items():
            if isinstance(prop, dict):
                example[str(name)] = _example_for_schema(prop, models, seen=set(seen))

    return example


def _body_fields_for_model(
    models: dict[str, Any], model_name: str, seen: set[str] | None = None
) -> list[dict[str, Any]]:
    if seen is None:
        seen = set()
    if not model_name or model_name in seen:
        return []
    seen.add(model_name)

    model = models.get(model_name)
    if not isinstance(model, dict):
        return []

    fields: list[dict[str, Any]] = []

    extends = model.get("extends")
    if isinstance(extends, dict):
        ref = extends.get("$ref")
        if isinstance(ref, str) and ref:
            fields.extend(_body_fields_for_model(models, ref, seen=set(seen)))

    required = _model_required_names(model)
    properties = model.get("properties")
    if not isinstance(properties, dict):
        return fields

    existing_names = {field["name"] for field in fields}
    for name, prop in properties.items():
        if not isinstance(prop, dict):
            continue
        field_name = str(name)
        if field_name in existing_names:
            continue
        fields.append(
            {
                "name": field_name,
                "type": _type_label_for_schema(prop),
                "required": field_name in required,
                "description": _clean_text(prop.get("description")),
            }
        )
        existing_names.add(field_name)

    return fields


def _extract_response_codes(operation: dict[str, Any]) -> list[str]:
    codes: list[str] = []
    for item in operation.get("responseMessages", []) or []:
        if not isinstance(item, dict):
            continue
        code = item.get("code")
        if code is None:
            continue
        message = _clean_text(item.get("message"))
        codes.append(f"{code} {message}".strip() if message else str(code))
    return _dedupe_texts(codes)


def _extract_response_content_types(
    operation: dict[str, Any], subdoc: dict[str, Any]
) -> list[str]:
    raw_types = operation.get("produces")
    if raw_types is None:
        raw_types = subdoc.get("produces")
    if isinstance(raw_types, str):
        raw_types = [raw_types]
    if not isinstance(raw_types, list):
        return []
    return _dedupe_texts([str(item) for item in raw_types if isinstance(item, str)])


def _extract_body_metadata(
    operation: dict[str, Any], models: dict[str, Any]
) -> tuple[Any | None, list[str], list[dict[str, Any]], str | None]:
    for param in operation.get("parameters", []) or []:
        if not isinstance(param, dict):
            continue
        if str(param.get("paramType") or "").lower() != "body":
            continue

        description = _clean_text(param.get("description"))
        model_name = None
        if isinstance(param.get("type"), str):
            model_name = param["type"]
        elif isinstance(param.get("$ref"), str):
            model_name = param["$ref"]
        elif isinstance(param.get("schema"), dict):
            schema = param["schema"]
            if isinstance(schema.get("$ref"), str):
                model_name = schema["$ref"]

        if model_name:
            fields = _body_fields_for_model(models, model_name)
            required = [field["name"] for field in fields if field.get("required")]
            example = _example_for_model(models, model_name)
            return example or None, required, fields, description

        field_name = str(param.get("name") or "body")
        field = {
            "name": field_name,
            "type": _type_label_for_schema(param),
            "required": bool(param.get("required")),
            "description": description,
        }
        example = {field_name: _example_for_schema(param, models)}
        required = [field_name] if field["required"] else []
        return example, required, [field], description

    return None, [], [], None


def _resolve_model_property_names(
    models: dict[str, Any], model_name: str, seen: set[str] | None = None
) -> list[str]:
    if not model_name:
        return []
    if seen is None:
        seen = set()
    if model_name in seen:
        return []
    seen.add(model_name)

    model = models.get(model_name)
    if not isinstance(model, dict):
        return []

    params: list[str] = []

    extends = model.get("extends")
    if isinstance(extends, dict):
        ref = extends.get("$ref")
        if isinstance(ref, str):
            params.extend(_resolve_model_property_names(models, ref, seen=seen))

    properties = model.get("properties")
    if isinstance(properties, dict):
        params.extend(str(name) for name in properties.keys())

    for submodels_key in ("subTypes", "subTypesModels"):
        submodels = model.get(submodels_key)
        if isinstance(submodels, list):
            for sub in submodels:
                if isinstance(sub, str):
                    params.extend(_resolve_model_property_names(models, sub, seen=seen))

    return _dedupe_keep_order(params)


def _extract_operation_params(
    operation: dict[str, Any], models: dict[str, Any]
) -> list[str]:
    params: list[str] = []
    for param in operation.get("parameters", []) or []:
        if not isinstance(param, dict):
            continue
        param_type = str(param.get("paramType") or "").lower()
        name = param.get("name")

        if param_type == "body":
            model_name = None
            if isinstance(param.get("type"), str):
                model_name = param["type"]
            elif isinstance(param.get("$ref"), str):
                model_name = param["$ref"]
            elif isinstance(param.get("schema"), dict):
                schema = param["schema"]
                if isinstance(schema.get("$ref"), str):
                    model_name = schema["$ref"]
                elif isinstance(schema.get("type"), str):
                    model_name = schema["type"]
            if model_name:
                params.extend(_resolve_model_property_names(models, model_name))
            elif isinstance(name, str) and name:
                params.append(name)
            continue

        if param_type == "path":
            continue

        if isinstance(name, str) and name:
            params.append(_normalize_param_name(name))

    return _dedupe_keep_order(params)


def _has_list_query_params(params: list[str]) -> bool:
    return any(param in _LIST_QUERY_PARAMS for param in params)


def _path_segments(path: str) -> list[str]:
    return [
        segment
        for segment in path.strip("/").split("/")
        if segment and segment != "api"
    ]


def _derive_service_key(base_path: str, normalized_path: str) -> str:
    base_segments = _path_segments(base_path)
    path_segments = _path_segments(normalized_path)
    if path_segments[: len(base_segments)] != base_segments:
        fixed = [
            segment
            for segment in path_segments
            if not (segment.startswith("{") and segment.endswith("}"))
        ]
        return fixed[-1] if fixed else "unknown"

    relative = path_segments[len(base_segments) :]
    if not relative:
        return base_segments[-1]

    if len(relative) == 1 and relative[0].startswith("{") and relative[0].endswith("}"):
        return base_segments[-1]
    if (
        len(relative) == 2
        and not relative[0].startswith("{")
        and relative[1].startswith("{")
        and relative[1].endswith("}")
    ):
        return base_segments[-1]

    fixed_suffix = [
        segment
        for segment in relative
        if not (segment.startswith("{") and segment.endswith("}"))
    ]
    if not fixed_suffix:
        return base_segments[-1]
    return base_segments[-1] + "-" + "-".join(fixed_suffix)


def _count_services(modules: dict[str, dict[str, Any]]) -> int:
    return sum(len(services or {}) for services in modules.values())


def _format_name_list(items: list[str], *, limit: int = 60) -> str:
    if not items:
        return "<none>"
    if len(items) <= limit:
        return ", ".join(items)
    shown = ", ".join(items[:limit])
    remaining = len(items) - limit
    return f"{shown}, ... (+{remaining} more)"


class ApiEndpointCache:
    def __init__(
        self,
        cp_client,
        *,
        token: str,
        cfg: EndpointCacheConfig | None = None,
        settings: Settings | None = None,
    ):
        self.cp = cp_client
        self.token = token
        self.cfg = cfg or EndpointCacheConfig()
        self.settings = settings or load_settings()
        self.cache_path = self.settings.paths.cache_dir / self.cfg.cache_filename
        self.settings.paths.ensure()

    def get_catalog(self, *, force_refresh: bool = False) -> dict[str, Any]:
        if not force_refresh:
            catalog = self._load_if_fresh()
            if catalog:
                return catalog
        catalog = self._build_catalog_from_clearpass()
        self._save(catalog)
        return catalog

    def _load_if_fresh(self) -> dict[str, Any] | None:
        try:
            stat = self.cache_path.stat()
        except FileNotFoundError:
            return None

        if time.time() - stat.st_mtime > self.cfg.ttl_seconds:
            return None

        try:
            data = json.loads(self.cache_path.read_text(encoding="utf-8"))
        except Exception:
            return None

        if (
            isinstance(data, dict)
            and data.get("version") in {2, 3}
            and isinstance(data.get("modules"), dict)
        ):
            return data
        return None

    def _save(self, api_catalog: dict[str, Any]) -> None:
        tmp = self.cache_path.with_suffix(".tmp")
        tmp.write_text(
            json.dumps(api_catalog, indent=2, sort_keys=True), encoding="utf-8"
        )
        os.replace(tmp, self.cache_path)

    def _raw_get_text(self, path: str) -> str:
        url = f"{self.cp.https_prefix}{self.cp.server}{path}"
        headers = {
            "Accept": "application/json, application/vnd.swagger+json, */*",
            "Authorization": f"Bearer {self.token}",
        }
        response = self.cp.session.get(
            url, headers=headers, verify=self.cp.verify_ssl, timeout=self.cp.timeout
        )
        response.raise_for_status()
        return response.text

    def _load_json(self, path: str) -> dict[str, Any] | None:
        try:
            text = self._raw_get_text(path)
        except Exception as exc:
            log.debug("[api_catalog] fetch %s failed: %s", path, exc)
            return None
        if not _is_json_content(text):
            return None
        try:
            parsed = json.loads(text)
        except Exception as exc:
            log.debug("[api_catalog] parse %s failed: %s", path, exc)
            return None
        return parsed if isinstance(parsed, dict) else None

    def _merge_action(
        self,
        service_actions: dict[str, Any],
        action_name: str,
        method: str,
        path: str,
        params: list[str],
        *,
        summary: str | None = None,
        notes: list[str] | None = None,
        response_codes: list[str] | None = None,
        response_content_types: list[str] | None = None,
        body_example: Any | None = None,
        body_required: list[str] | None = None,
        body_fields: list[dict[str, Any]] | None = None,
        body_description: str | None = None,
    ) -> None:
        entry = service_actions.setdefault(action_name, {"method": method, "paths": []})
        entry["method"] = method
        entry["paths"] = _dedupe_keep_order(entry.get("paths", []) + [path])
        if params:
            entry["params"] = _dedupe_keep_order(entry.get("params", []) + params)
        if summary and not entry.get("summary"):
            entry["summary"] = summary
        if notes:
            entry["notes"] = _dedupe_texts(entry.get("notes", []) + notes)
        if response_codes:
            entry["response_codes"] = _dedupe_texts(
                entry.get("response_codes", []) + response_codes
            )
        if response_content_types:
            entry["response_content_types"] = _dedupe_texts(
                entry.get("response_content_types", []) + response_content_types
            )
        if body_description and not entry.get("body_description"):
            entry["body_description"] = body_description
        if body_required:
            entry["body_required"] = _dedupe_keep_order(
                entry.get("body_required", []) + body_required
            )
        if body_fields:
            existing_names = {
                field.get("name")
                for field in entry.get("body_fields", [])
                if isinstance(field, dict)
            }
            merged_fields = list(entry.get("body_fields", []))
            for field in body_fields:
                if not isinstance(field, dict):
                    continue
                name = field.get("name")
                if name in existing_names:
                    continue
                merged_fields.append(field)
                existing_names.add(name)
            entry["body_fields"] = merged_fields
        if (
            body_example not in (None, {}, [])
            and entry.get("body_example") in (None, {}, [])
        ):
            entry["body_example"] = body_example

    def _process_swagger_subdoc(
        self, module_services: dict[str, Any], subdoc: dict[str, Any]
    ) -> None:
        apis = subdoc.get("apis")
        if not isinstance(apis, list) or not apis:
            return

        models = subdoc.get("models")
        if not isinstance(models, dict):
            models = {}

        resource_path = subdoc.get("resourcePath") or apis[0].get("path")
        if not isinstance(resource_path, str):
            return
        base_path = _normalize_template_placeholders(
            _ensure_api_prefix(_clean_apigility_route(resource_path))
        )

        raw_entries: list[dict[str, Any]] = []
        for api_item in apis:
            if not isinstance(api_item, dict):
                continue
            raw_path = api_item.get("path")
            if not isinstance(raw_path, str):
                continue
            normalized_path = _normalize_template_placeholders(
                _ensure_api_prefix(raw_path)
            )
            service_key = _derive_service_key(base_path, normalized_path)
            for operation in api_item.get("operations", []) or []:
                if not isinstance(operation, dict):
                    continue
                method = str(operation.get("method") or "").upper().strip()
                if method not in {"GET", "POST", "DELETE", "PATCH", "PUT"}:
                    continue
                params = _extract_operation_params(operation, models)
                body_example, body_required, body_fields, body_description = (
                    _extract_body_metadata(operation, models)
                )
                raw_entries.append(
                    {
                        "service": service_key,
                        "base_path": base_path,
                        "path": normalized_path,
                        "method": method,
                        "params": params,
                        "placeholders": _extract_placeholders(normalized_path),
                        "summary": _clean_text(operation.get("summary")),
                        "notes": _dedupe_texts(
                            [_clean_text(operation.get("notes")) or ""]
                        ),
                        "response_codes": _extract_response_codes(operation),
                        "response_content_types": _extract_response_content_types(
                            operation, subdoc
                        ),
                        "body_example": body_example,
                        "body_required": body_required,
                        "body_fields": body_fields,
                        "body_description": body_description,
                    }
                )

        by_service: dict[str, list[dict[str, Any]]] = {}
        for entry in raw_entries:
            by_service.setdefault(entry["service"], []).append(entry)

        for service_key, entries in by_service.items():
            service = module_services.setdefault(service_key, {"actions": {}})
            actions = service.setdefault("actions", {})
            has_item_style = any(entry["placeholders"] for entry in entries)
            has_post_on_base = any(
                entry["method"] == "POST" and entry["path"] == entry["base_path"]
                for entry in entries
            )

            for entry in entries:
                method = entry["method"]
                params = entry["params"]
                path = entry["path"]
                base_path = entry["base_path"]

                if method == "GET":
                    is_base_get = path == base_path
                    if is_base_get and (
                        _has_list_query_params(params)
                        or (
                            not entry["placeholders"]
                            and (has_item_style or has_post_on_base)
                        )
                    ):
                        action_name = "list"
                    else:
                        action_name = "get"
                elif method == "POST":
                    action_name = "add"
                elif method == "DELETE":
                    action_name = "delete"
                elif method == "PATCH":
                    action_name = "update"
                else:
                    action_name = "replace"

                self._merge_action(
                    actions,
                    action_name,
                    method,
                    path,
                    params,
                    summary=entry.get("summary"),
                    notes=entry.get("notes"),
                    response_codes=entry.get("response_codes"),
                    response_content_types=entry.get("response_content_types"),
                    body_example=entry.get("body_example"),
                    body_required=entry.get("body_required"),
                    body_fields=entry.get("body_fields"),
                    body_description=entry.get("body_description"),
                )

    def _process_apigility_services(
        self, module_services: dict[str, Any], listing: dict[str, Any]
    ) -> bool:
        services = listing.get("services")
        if not isinstance(services, list) or not services:
            return False

        added = 0
        for service in services:
            if not isinstance(service, dict):
                continue
            route = service.get("route")
            if not isinstance(route, str):
                continue

            base_path = _normalize_template_placeholders(
                _ensure_api_prefix(_clean_apigility_route(route))
            )
            service_key = _camel_to_kebab(
                str(service.get("name") or base_path.strip("/").split("/")[-1])
            )
            service_entry = module_services.setdefault(service_key, {"actions": {}})
            actions = service_entry.setdefault("actions", {})

            collection_methods = {
                str(method).upper()
                for method in service.get("collection_http_methods", [])
                if isinstance(method, str)
            }
            entity_methods = {
                str(method).upper()
                for method in service.get("entity_http_methods", [])
                if isinstance(method, str)
            }
            entity_id = str(service.get("entity_identifier_name") or "id")
            entity_id = _normalize_param_name(entity_id)
            entity_path = base_path + "/{" + entity_id + "}"

            for method in sorted(collection_methods):
                if method == "GET":
                    self._merge_action(
                        actions, "list", "GET", base_path, list(_LIST_QUERY_PARAMS)
                    )
                elif method == "POST":
                    self._merge_action(actions, "add", "POST", base_path, [])

            for method in sorted(entity_methods):
                if method == "GET":
                    self._merge_action(actions, "get", "GET", entity_path, [])
                elif method == "DELETE":
                    self._merge_action(actions, "delete", "DELETE", entity_path, [])
                elif method == "PATCH":
                    self._merge_action(actions, "update", "PATCH", entity_path, [])
                elif method == "PUT":
                    self._merge_action(actions, "replace", "PUT", entity_path, [])

            added += 1

        return added > 0

    def _log_module_services(
        self, cli_module: str, module_services: dict[str, Any]
    ) -> None:
        service_names = sorted(module_services.keys())
        if service_names:
            log.debug(
                "[api_catalog] Loaded module: %s with (%d) services -> %s",
                cli_module,
                len(service_names),
                _format_name_list(service_names),
            )
        else:
            log.info("[api_catalog] %s: 0 services", cli_module)

    def _build_catalog_from_clearpass(self) -> dict[str, Any]:
        api_docs_text = self._raw_get_text("/api-docs")
        module_doc_paths = _extract_modules_from_api_docs(api_docs_text)

        log.info("Discovered %d modules from /api-docs", len(module_doc_paths))
        discovered_modules = [
            _module_to_cli(path.rsplit("/", 1)[-1]) for path in module_doc_paths
        ]
        if discovered_modules:
            log.debug(
                "[api_catalog] modules: %s", _format_name_list(discovered_modules)
            )

        modules: dict[str, dict[str, Any]] = {}

        for module_path in module_doc_paths:
            module_name = module_path.rsplit("/", 1)[-1]
            cli_module = _module_to_cli(module_name)
            module_services = modules.setdefault(cli_module, {})

            listing = self._load_json(f"/api/apigility/documentation/{module_name}")
            if listing is None:
                for alt in (
                    f"/api/apigility/documentation/{module_name}/swagger",
                    module_path,
                    f"{module_path}.json",
                ):
                    listing = self._load_json(alt)
                    if listing is not None:
                        break

            if listing is None:
                log.warning("[api_catalog] %s: no JSON docs found", module_name)
                continue

            if self._process_apigility_services(module_services, listing):
                self._log_module_services(cli_module, module_services)
                continue

            listing_apis = listing.get("apis")
            if not isinstance(listing_apis, list) or not listing_apis:
                log.warning(
                    "[api_catalog] %s: JSON docs contained neither usable "
                    "'services' nor usable 'apis' (keys=%s)",
                    module_name,
                    list(listing.keys())[:30],
                )
                continue

            for item in listing_apis:
                if not isinstance(item, dict):
                    continue
                path = item.get("path")
                if not isinstance(path, str):
                    continue

                if path.startswith("/api/"):
                    sub_path = path
                elif path.startswith(f"/{module_name}/"):
                    sub_path = f"/api/apigility/documentation{path}"
                elif path.startswith("/"):
                    sub_path = f"/api/apigility/documentation/{module_name}{path}"
                else:
                    sub_path = f"/api/apigility/documentation/{module_name}/{path}"

                subdoc = self._load_json(sub_path)
                if not subdoc:
                    continue
                self._process_swagger_subdoc(module_services, subdoc)

            if not module_services:
                log.warning("[api_catalog] %s: no services were extracted", module_name)
                continue

            self._log_module_services(cli_module, module_services)

        catalog: dict[str, Any] = {
            "version": 3,
            "generated_at": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
            "server": getattr(self.cp, "server", None),
            "modules": modules,
        }
        log.info("Total modules in cache: %d", len(modules))
        log.info("Total services in cache: %d", _count_services(modules))
        return catalog


def get_api_catalog(
    cp_client,
    *,
    token: str,
    force_refresh: bool = False,
    settings: Settings | None = None,
) -> dict[str, Any]:
    return ApiEndpointCache(cp_client, token=token, settings=settings).get_catalog(
        force_refresh=force_refresh
    )


def get_cache_file_path(settings: Settings | None = None) -> Path:
    cfg = EndpointCacheConfig()
    active_settings = settings or load_settings()
    active_settings.paths.ensure()
    return active_settings.paths.cache_dir / cfg.cache_filename


def load_cached_catalog(settings: Settings | None = None) -> dict[str, Any] | None:
    path = get_cache_file_path(settings=settings)
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
    except FileNotFoundError:
        return None
    except Exception:
        return None

    if (
        isinstance(data, dict)
        and data.get("version") in {2, 3}
        and isinstance(data.get("modules"), dict)
    ):
        return data
    return None


def clear_api_cache(settings: Settings | None = None) -> bool:
    path = get_cache_file_path(settings=settings)
    try:
        path.unlink()
        return True
    except FileNotFoundError:
        return False



from __future__ import annotations


def _normalize_flag_name(key: str) -> str:
    return key.replace("-", "_")


def parse_cli(argv: list[str]) -> dict:
    args: dict[str, object] = {}
    positionals: list[str] = []
    completing = "--_complete" in argv

    boolean_flags = {
        "verbose",
        "version",
        "debug",
        "console",
        "all",
        "decrypt",
        "dry_run",
        "continue_on_error",
        "help",
    }

    for item in argv[1:]:
        if item == "--_complete":
            args["_complete"] = True
        elif item.startswith("--_cword="):
            args["_cword"] = int(item.split("=", 1)[1])
        elif item.startswith("--_cur="):
            args["_cur"] = item.split("=", 1)[1]
        elif item in {"?", "-h", "--help"}:
            args["help"] = True
        elif item == "--":
            continue
        elif item.startswith("--") and "=" in item:
            key, value = item[2:].split("=", 1)
            args[_normalize_flag_name(key)] = value
        elif item.startswith("--"):
            key = _normalize_flag_name(item[2:])
            if key in boolean_flags:
                args[key] = True
            elif completing:
                continue
            else:
                raise ValueError(f"Unknown flag: {item}")
        elif item.startswith("-"):
            if completing:
                continue
            raise ValueError(f"Unknown flag: {item}")
        else:
            positionals.append(item)

    if len(positionals) >= 1:
        args["module"] = positionals[0]
    if positionals[:1] == ["copy"]:
        if len(positionals) >= 2:
            args["copy_module"] = positionals[1]
        if len(positionals) >= 3:
            args["copy_service"] = positionals[2]
    else:
        if len(positionals) >= 2:
            args["service"] = positionals[1]
        if len(positionals) >= 3:
            args["action"] = positionals[2]

    return args



from __future__ import annotations

from copy import deepcopy
from typing import Any

from arapy.core.resolver import query_params_for_action

DEFAULT_PAGE_SIZE = 1000


def _extract_items(response: Any) -> list[Any] | None:
    if isinstance(response, dict):
        embedded = response.get("_embedded")
        if isinstance(embedded, dict):
            items = embedded.get("items")
            if isinstance(items, list):
                return items
        return None
    if isinstance(response, list):
        return response
    return None


def _extract_total_count(response: Any) -> int | None:
    if not isinstance(response, dict):
        return None
    count = response.get("count")
    if isinstance(count, bool):
        return None
    if isinstance(count, int):
        return count
    if isinstance(count, str) and count.isdigit():
        return int(count)
    return None


def _merge_list_responses(
    first_response: Any,
    all_items: list[Any],
    *,
    total_count: int | None,
) -> Any:
    if isinstance(first_response, dict):
        merged = deepcopy(first_response)
        embedded = merged.get("_embedded")
        if isinstance(embedded, dict) and isinstance(embedded.get("items"), list):
            embedded["items"] = all_items
        if total_count is not None:
            merged["count"] = total_count
        elif "count" in merged:
            merged["count"] = len(all_items)
        links = merged.get("_links")
        if isinstance(links, dict) and "next" in links:
            links = dict(links)
            links.pop("next", None)
            merged["_links"] = links
        return merged
    if isinstance(first_response, list):
        return all_items
    return first_response


def fetch_all_list_results(cp, token: str, api_catalog: dict, args: dict[str, Any]):
    params = query_params_for_action(cp, api_catalog, args, "list")
    action_def = cp.get_action_definition(
        api_catalog, args["module"], args["service"], "list"
    )
    allowed = {
        str(name) for name in action_def.get("params", []) or [] if isinstance(name, str)
    }

    explicit_limit = "limit" in args and args.get("limit") not in (None, "")
    if "limit" in allowed and not explicit_limit:
        params["limit"] = DEFAULT_PAGE_SIZE
    if "offset" in allowed and "offset" not in params:
        params["offset"] = 0

    response = cp.list(api_catalog, token, args, params=params or None)
    page_items = _extract_items(response)
    if page_items is None or "limit" not in allowed or "offset" not in allowed:
        return response
    if explicit_limit:
        return response

    all_items = list(page_items)
    total_count = _extract_total_count(response)
    page_size = int(params["limit"])
    current_offset = int(params.get("offset", 0))

    while True:
        if total_count is not None and len(all_items) >= total_count:
            break
        if len(page_items) < page_size:
            break

        next_offset = current_offset + len(page_items)
        if next_offset <= current_offset:
            break

        next_params = dict(params)
        next_params["offset"] = next_offset
        if "calculate_count" in next_params and total_count is not None:
            next_params["calculate_count"] = "false"

        next_response = cp.list(api_catalog, token, args, params=next_params or None)
        page_items = _extract_items(next_response)
        if page_items is None or not page_items:
            break

        all_items.extend(page_items)
        current_offset = next_offset

        if total_count is None:
            total_count = _extract_total_count(next_response)

    return _merge_list_responses(response, all_items, total_count=total_count)



from __future__ import annotations

import csv
import json
import logging
from pathlib import Path
from typing import Any, Iterable

from arapy.core.config import SECRET_FIELDS, Settings
from arapy.io.files import ensure_parent_dir

log = logging.getLogger(__name__)


def should_mask_secrets(args: dict | None, settings: Settings) -> bool:
    if not args:
        return settings.encrypt_secrets

    if args.get("decrypt"):
        return False

    encrypt = args.get("encrypt")
    if encrypt is None:
        return settings.encrypt_secrets

    value = str(encrypt).strip().lower()
    if value in {"enable", "enabled", "true", "1", "yes", "on"}:
        return True
    if value in {"disable", "disabled", "false", "0", "no", "off"}:
        return False
    raise ValueError("--encrypt must be enable or disable")


def sanitize_secrets(value: Any, *, mask_secrets: bool = True):
    if not mask_secrets:
        return value

    if isinstance(value, dict):
        sanitized = {}
        for key, item in value.items():
            if key in SECRET_FIELDS:
                sanitized[key] = ""
            else:
                sanitized[key] = sanitize_secrets(item, mask_secrets=mask_secrets)
        return sanitized

    if isinstance(value, list):
        return [sanitize_secrets(item, mask_secrets=mask_secrets) for item in value]

    return value


def _extract_by_path(data: Any, path: Iterable[str | int]):
    current = data
    for step in path:
        try:
            current = current[step]
        except (KeyError, IndexError, TypeError):
            return None
    return current


def _console_text_for_raw_bytes(value: bytes) -> str:
    try:
        decoded = value.decode("utf-8")
    except UnicodeDecodeError:
        return f"<binary data: {len(value)} bytes>"

    if any(ord(char) < 32 and char not in "\r\n\t" for char in decoded):
        return f"<binary data: {len(value)} bytes>"

    if decoded and not any(char.isprintable() and not char.isspace() for char in decoded):
        return f"<binary data: {len(value)} bytes>"

    return decoded


def write_value_to_file(
    value: Any,
    path: str | Path,
    *,
    mode: str = "w",
    data_format: str = "json",
    csv_fieldnames: list[str] | None = None,
    csv_include_header: bool = True,
    items_path: tuple[str | int, ...] = ("_embedded", "items"),
    also_console: bool = False,
    mask_secrets: bool = True,
) -> None:
    if mode not in {"a", "w"}:
        raise ValueError("mode must be 'a' or 'w'")
    if data_format not in {"json", "csv", "raw"}:
        raise ValueError("data_format must be 'json', 'csv', or 'raw'")

    path = Path(path)
    ensure_parent_dir(path)
    safe_value = sanitize_secrets(value, mask_secrets=mask_secrets)

    if data_format == "json":
        with path.open(mode, encoding="utf-8") as handle:
            if isinstance(safe_value, (dict, list)):
                json.dump(safe_value, handle, indent=2, ensure_ascii=False)
                handle.write("\n")
            else:
                handle.write(f"{safe_value}\n")
        if also_console:
            if isinstance(safe_value, (dict, list)):
                print(json.dumps(safe_value, indent=2, ensure_ascii=False))
            else:
                print(safe_value)
    elif data_format == "raw":
        if isinstance(safe_value, bytes):
            binary_mode = "ab" if mode == "a" else "wb"
            with path.open(binary_mode) as handle:
                handle.write(safe_value)
            if also_console:
                print(_console_text_for_raw_bytes(safe_value))
        else:
            text = str(safe_value)
            with path.open(mode, encoding="utf-8") as handle:
                handle.write(text)
            if also_console:
                print(text)
    else:
        rows = None
        if isinstance(safe_value, dict):
            extracted = _extract_by_path(safe_value, items_path)
            if isinstance(extracted, list):
                rows = extracted
        if rows is None and isinstance(safe_value, list):
            rows = safe_value
        if rows is None and isinstance(safe_value, dict):
            rows = [safe_value]
        if rows is None:
            rows = [{"value": safe_value}]
        if not rows:
            return

        append_mode = mode == "a"
        need_header = csv_include_header and (
            not append_mode
            or (append_mode and (not path.exists() or path.stat().st_size == 0))
        )

        with path.open(mode, encoding="utf-8", newline="") as handle:
            if isinstance(rows[0], dict):
                fieldnames = csv_fieldnames or list(rows[0].keys())
                writer = csv.DictWriter(
                    handle,
                    fieldnames=fieldnames,
                    lineterminator="\n",
                    extrasaction="ignore",
                )
                if need_header:
                    writer.writeheader()
                    if also_console:
                        print(",".join(fieldnames))
                for row in rows:
                    writer.writerow(row)
                    if also_console:
                        print(
                            ",".join(
                                "" if row.get(field) is None else str(row.get(field))
                                for field in fieldnames
                            )
                        )
            else:
                writer = csv.writer(handle, lineterminator="\n")
                if need_header:
                    writer.writerow(["value"])
                    if also_console:
                        print("value")
                for row in rows:
                    writer.writerow([row])
                    if also_console:
                        print(row)

    log.debug("Wrote file to %s", path)


def log_to_file(
    thing: Any,
    filename: str | Path | None = None,
    *args,
    also_console: bool = False,
    mode: str = "w",
    data_format: str = "json",
    csv_fieldnames: list[str] | None = None,
    csv_include_header: bool = True,
    items_path: tuple[str | int, ...] = ("_embedded", "items"),
    mask_secrets: bool = True,
    **kwargs,
):
    if filename is None:
        raise ValueError("filename must be provided")

    if callable(thing):
        result = thing(*args, **kwargs)
        if result is not None:
            write_value_to_file(
                result,
                filename,
                mode=mode,
                data_format=data_format,
                csv_fieldnames=csv_fieldnames,
                csv_include_header=csv_include_header,
                items_path=items_path,
                also_console=also_console,
                mask_secrets=mask_secrets,
            )
        return result

    write_value_to_file(
        thing,
        filename,
        mode=mode,
        data_format=data_format,
        csv_fieldnames=csv_fieldnames,
        csv_include_header=csv_include_header,
        items_path=items_path,
        also_console=also_console,
        mask_secrets=mask_secrets,
    )
    return thing



from __future__ import annotations

import sys
from dataclasses import replace

import urllib3

from arapy import get_version
from arapy.cli.commands import ACTIONS
from arapy.cli.completion import print_completions
from arapy.cli.copy import handle_copy_command
from arapy.cli.help import render_help
from arapy.cli.parser import parse_cli
from arapy.cli.server import handle_server_command
from arapy.core.catalog import (
    OAUTH_ENDPOINTS,
    clear_api_cache,
    get_api_catalog,
    load_cached_catalog,
)
from arapy.core.client import ClearPassClient
from arapy.core.config import Settings, load_settings
from arapy.io.files import load_api_token_file
from arapy.io.output import should_mask_secrets
from arapy.logging.setup import LOG_LEVELS, configure_logging


def build_client(settings: Settings, *, mask_secrets: bool = True) -> ClearPassClient:
    if not settings.server:
        raise ValueError(
            "ARAPY_SERVER is not configured. Set it in the "
            "environment before running network actions."
        )
    try:
        return ClearPassClient(
            server=settings.server,
            https_prefix=settings.https_prefix,
            verify_ssl=settings.verify_ssl,
            timeout=settings.timeout,
            mask_secrets=mask_secrets,
        )
    except TypeError as exc:
        if "mask_secrets" not in str(exc):
            raise
        cp = ClearPassClient(
            server=settings.server,
            https_prefix=settings.https_prefix,
            verify_ssl=settings.verify_ssl,
            timeout=settings.timeout,
        )
        setattr(cp, "mask_secrets", mask_secrets)
        return cp


def print_help(args: dict | None = None) -> None:
    catalog = load_cached_catalog()
    print(render_help(catalog, args or {}, version=get_version()))


def complete(words: list[str], settings: Settings | None = None) -> None:
    catalog = load_cached_catalog(settings=settings)
    print_completions(words, catalog)


def settings_with_cli_overrides(settings: Settings, args: dict) -> Settings:
    api_token = args.get("api_token") or args.get("token") or settings.api_token
    token_file = (
        args.get("token_file")
        or args.get("api_token_file")
        or settings.api_token_file
    )
    return replace(settings, api_token=api_token, api_token_file=token_file)


def resolve_auth_token(cp: ClearPassClient, settings: Settings) -> str:
    if settings.api_token:
        return settings.api_token
    if settings.api_token_file:
        return load_api_token_file(settings.api_token_file)
    return cp.login(OAUTH_ENDPOINTS, settings.credentials)["access_token"]


def main() -> None:
    settings = load_settings()
    if not settings.verify_ssl:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    if "--_complete" in sys.argv:
        words = [word for word in sys.argv[1:] if word != "--_complete"]
        complete(words, settings=settings)
        return

    log_mgr = configure_logging(settings, root_name="arapy")
    log = log_mgr.get_logger(__name__)

    args = parse_cli(sys.argv)
    active_settings = settings_with_cli_overrides(settings, args)

    log_level = args.get("log_level")
    if log_level:
        normalized = str(log_level).upper()
        if normalized not in LOG_LEVELS:
            valid = ", ".join(name.lower() for name in LOG_LEVELS)
            log.error("Invalid log level: %s. Valid options are: %s", log_level, valid)
            return
        log_mgr.set_level(LOG_LEVELS[normalized])

    if args.get("version"):
        print(get_version())
        return

    if args.get("help"):
        print_help(args)
        return

    if not args.get("module"):
        print_help({})
        return

    if args.get("module") == "cache":
        service = args.get("service")
        if service == "clear" and not args.get("action"):
            removed = clear_api_cache(settings=active_settings)
            if removed:
                log.info("API endpoint cache cleared.")
            else:
                log.info("No API endpoint cache file found (already clear).")
            return
        if service == "update" and not args.get("action"):
            cp = build_client(active_settings)
            token = resolve_auth_token(cp, active_settings)
            get_api_catalog(
                cp, token=token, force_refresh=True, settings=active_settings
            )
            return
        print_help({"module": "cache"})
        return

    if args.get("module") == "server":
        if handle_server_command(args):
            return
        print_help({"module": "server", "service": args.get("service")})
        return

    if args.get("module") == "copy":
        handle_copy_command(
            args,
            settings=active_settings,
            build_client=build_client,
            resolve_auth_token=resolve_auth_token,
            get_api_catalog=get_api_catalog,
        )
        return

    module = args.get("module")
    service = args.get("service")
    action = args.get("action")

    if not (module and service and action):
        print_help(args)
        return

    try:
        command = ACTIONS[action]
    except KeyError:
        print_help(args)
        print(f"\nUnknown command: {module} {service} {action}")
        return

    mask_secrets = should_mask_secrets(args, active_settings)
    cp = build_client(active_settings, mask_secrets=mask_secrets)
    log.info(
        "Connecting to ClearPass server: %s (SSL verify: %s)",
        active_settings.server,
        active_settings.verify_ssl,
    )
    token = resolve_auth_token(cp, active_settings)
    log.debug(f"Session token: {token}")
    api_catalog = get_api_catalog(cp, token=token, settings=active_settings)
    command(cp, token, api_catalog, args, settings=active_settings)
